/*******************************************************************************

    cba_mem.c
    Copyright 2017-2020 NetApp, Inc.

LICENSE:
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include <config-util.h>
#include <grub/env.h>
#include <grub/file.h>
#include <grub/dl.h>
#include <grub/command.h>

GRUB_MOD_LICENSE ("GPLv3+");
GRUB_MOD_DUAL_LICENSE ("All Rights Reserved");

#define PARAM_FILE  "(lvm/system-cba.active)/cba.cfg"

#define DBG_PRT(arg...)    \
{                          \
    if (enableDbgPrints)   \
        grub_printf(arg);  \
}

static int enableDbgPrints = 0;

static const char memParamString[]="MEMORY=";
static const char coreParamString[]="CORES=";

/**
 * This function opens a the CBA parameter file, scans
 * it for the memory request, and returns it if present.
 */
static int
grub_read_cfg(char *memsz_str, int memsz_len, char *cpus_str, int cpus_len)
{
    grub_file_t file;
    char*       read_buf;
    char*       read_buf_end;
    char*       buf_ptr;
    grub_off_t  file_len;
    unsigned long mem_req = 0;
    unsigned long mem_param_hdr_size;
    unsigned long cpu_req = 0;
    unsigned long cpu_param_hdr_size;

    /* Read in file to a temporary buffer */
    file = grub_file_open(PARAM_FILE);
    if (!file)
    {
        DBG_PRT("Could not open file: %s\n", PARAM_FILE);
        return -1;
    }

    read_buf = (char *)grub_zalloc(file->size);
    if (!read_buf)
    {
        grub_printf("Allocation failure %" PRIuGRUB_UINT64_T " bytes)\n", file->size);
        grub_file_close(file);
        return -1;
    }

    file_len = grub_file_read(file, read_buf, file->size);
    if (file_len != file->size)
    {
        grub_printf("Failed to read file (read return 0x%" PRIxGRUB_UINT64_T " on attempted read of %" PRIuGRUB_UINT64_T ")\n", file_len, file->size);
        grub_free(read_buf);
        grub_file_close(file);
        return -1;
    }
    grub_file_close(file);

    mem_param_hdr_size = grub_strlen(memParamString);
    cpu_param_hdr_size = grub_strlen(coreParamString);
    read_buf_end = read_buf + file_len;
    buf_ptr = read_buf;

    while (buf_ptr < read_buf_end)
    {
        if (*buf_ptr != '#')    /* Skip coment lines. */
        {
            if (grub_strncmp(buf_ptr, memParamString, mem_param_hdr_size) == 0)
            {
                DBG_PRT("Found Mem header\n");
                buf_ptr += mem_param_hdr_size;
                mem_req = grub_strtoul(buf_ptr, 0, 10);
            }
            else if (grub_strncmp(buf_ptr, coreParamString, cpu_param_hdr_size) == 0)
            {
                DBG_PRT("Found cores header\n");
                buf_ptr += cpu_param_hdr_size;
                cpu_req = grub_strtoul(buf_ptr, 0, 10);
            }
        }

        if (mem_req && cpu_req)
            break;

        /* Advance to the next new line. */
        while (buf_ptr < read_buf_end && *buf_ptr != '\n')
            ++buf_ptr;

        /* Consume the new line. */
        ++buf_ptr;
    }

    /* If the string parsing didn't find a valid number above, the
     * strtoul conversion would return a 0, which is what we would
     * want by default.
     */
    grub_snprintf(memsz_str, memsz_len, "%lu", mem_req);
    grub_snprintf(cpus_str, cpus_len, "%lu", cpu_req);
    grub_free(read_buf);
    return 0;
}


/**
 * This function implements the cba_mem grub command.  It scans
 * the CBA parameters file for the requested memory and cores,
 * then creates/sets the corresponding grub environment variables
 * with the values. If the parameter file does not exist, it will
 * set the variable to 0, assuming no CBA memory or cores are required.
 */
static grub_err_t
grub_cmd_cba_mem
    (
    grub_command_t cmd __attribute__ ((unused)),
    int argc,
    char *argv[]
    )
{
    char memVal[80];
    char cpuVal[80];
    char *memVar;
    char *cpuVar;
    const char *defaultValue = "0";

    if (argc != 2)
    {
        return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("two arguments expected"));
    }

    /* If "DEBUG" env var exists, enable prints */
    enableDbgPrints = grub_env_get("DEBUG") ? 1 : 0;

    memVar = argv[0];
    cpuVar = argv[1];

    /* Get memory configuration */
    DBG_PRT("variables: %s,%s\n", memVar, cpuVar);

    if (grub_read_cfg(memVal, sizeof(memVal), cpuVal, sizeof(cpuVal)))
    {
        /* If it failed to read the value from config file, then assign the default value */
        grub_strncpy(memVal, defaultValue, sizeof(memVal));
        grub_strncpy(cpuVal, defaultValue, sizeof(cpuVal));

        /* Ignore the error to avoid waiting for user response during boot */
        grub_errno = 0;
    }

    DBG_PRT("Set %s=%s\n", memVar, memVal);
    grub_env_set(memVar, memVal);

    DBG_PRT("Set %s=%s\n", cpuVar, cpuVal);
    grub_env_set(cpuVar, cpuVal);

    return GRUB_ERR_NONE;
}

static grub_command_t cmd;

/* Register/unregister our command with module init/exit */
GRUB_MOD_INIT(cba_mem)
{
    cmd = grub_register_command("cba_mem", grub_cmd_cba_mem, N_("ENV_VAR ENV_VAR"),
                                N_("Provides memory in GB and CPU cores requested for CBA."));
}

GRUB_MOD_FINI(cba_mem)
{
    if (cmd)
    {
        grub_unregister_command(cmd);
        cmd = 0;
    }
}
