/*******************************************************************************

    cpu_map.c
    Copyright 2016-2020 NetApp, Inc.

DESCRIPTION: Grub2 module to detect number of logical cores present and set
isolcpus kernel configuration parameter accordingly. The current implementation
makes core 0 and any container cores, including their hyperthread cores,
available to the scheduler and reserves remaining cores for RAID application.

LICENSE:
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

/***  INCLUDES  ***/
#include <grub/dl.h>
#include <grub/env.h>
#include <grub/file.h>
#include <grub/command.h>

/***  CONSTANT DEFINITIONS  ***/
GRUB_MOD_LICENSE ("GPLv3+");
GRUB_MOD_DUAL_LICENSE ("All Rights Reserved");

#define CPUID_LEAF (0xB)
#define NUM_LOGICAL_CORES_MASK (0xFFFF)
#define LEVEL_NUM_MASK (0xFF)

#define CPU_CONFIG_FILE "(lvm/system-raid.active)/conf/platResAlloc.cfg"

/*** MACRO DEFINITIONS ***/
#define DBG_PRT(arg...)  do { if (enableDbgPrints) grub_printf(arg); } while (0)


/***  TYPE DEFINITIONS  ***/
typedef struct {
    char*       start;
    char*       end;
    char*       cur;
    grub_off_t  len;
} ConfigFile;


/***  LOCALS  ***/
static int enableDbgPrints = 0;

static const char coreConfigString[]="CPULEOS|";
static const char normalModeString[]="Norm|";
static const char eosDomainString[]="EOS|";

/*** PROCEDURES ***/

/**
 * Takes a pointer and returns a pointer to the next new line (or NULL if there
 * are no new lines remaining).
 */
static char*
getNextString(ConfigFile* configFile)
{
    char* cur = configFile->cur + grub_strlen(configFile->cur);

    while (cur < configFile->end && *cur == '\0')
        cur++;

    configFile->cur = cur;

    if (cur < configFile->end)
        return cur;

    return 0;
}

/**
 * Free up memory allocated when file was open and read
 */
static void
configFileFree(ConfigFile* configFile)
{
    grub_free(configFile->start);
    configFile->start = 0;
}

/**
 * Read in the entire configuration file and store in structure.
 * Also convert all newlines to string terminators so that it is easier to parse.
 */
static int
readConfigFile(ConfigFile *configFile)
{
    grub_file_t  file;
    char        *start;

    file = grub_file_open(CPU_CONFIG_FILE);
    if (!file)
    {
        grub_printf("Could not open config file: %s\n",
                    CPU_CONFIG_FILE);

        /* Ignore the error to avoid waiting for user response during boot */
        grub_errno = 0;
        return 0;
    }

    configFile->start = (char *)grub_zalloc(file->size);
    if (!configFile->start)
    {
        grub_file_close(file);
        return 0;
    }

    configFile->len = grub_file_read(file, configFile->start, file->size);
    if (configFile->len != file->size)
    {
        configFileFree(configFile);
        grub_file_close(file);
        return 0;
    }
    grub_file_close(file);

    configFile->cur = configFile->start;
    configFile->end = configFile->start + configFile->len;
    start = configFile->start;

    /* Scrub the file to get rid of newlines, replacing them with null terminator so that */
    /* parsing is easier */
    while (start < configFile->end)
    {
        if (*start == '\r' || *start == '\n')
            *start = '\0';
        start++;
    }

    return 1;
}

/**
 * Takes the actual/physical mem size and looks for a corresponding memory
 * mapping size in the config file. Returns 0 if no valid config line found.
 * If it returns mapping, that memory will be allocated and needs to be freed
 * by the caller.
        RETURN number of CBA cores to be used.
 */
static unsigned long
getCpuConfiguration(unsigned long cbaCoresReq)
{
    ConfigFile configFile;
    char *cur;
    unsigned long cpu_param_hdr_size;
    unsigned long dom_field_size;
    unsigned long mode_field_size;
    unsigned long core_max = 0;
    unsigned long core_min;

    DBG_PRT("cbaCoresReq: %lu\n", cbaCoresReq);

    if (!readConfigFile(&configFile))
    {
        return 0;
    }

    cur = getNextString(&configFile);
    cpu_param_hdr_size = grub_strlen(coreConfigString);
    dom_field_size = grub_strlen(eosDomainString);
    mode_field_size = grub_strlen(normalModeString);

    while (cur != NULL && *cur != '\0')
    {
        /* Advance to the next non-white space character */
        while((*cur != '\0') && grub_isspace(*cur))
        {
            cur++;
        }

        /* If comment line, then skip to next line in file */
        if (( *cur == '#') || (grub_strncmp(cur, coreConfigString, cpu_param_hdr_size) != 0))
        {
            cur = getNextString(&configFile);
            continue;
        }

        cur += cpu_param_hdr_size;

        /* Advance to Domain */
        while(*cur != '|')
        {
            cur++;
        }
        cur++;
        if (grub_strncmp(cur, eosDomainString, dom_field_size) != 0)
        {
            cur = getNextString(&configFile);
            continue;
        }
        cur += dom_field_size;

        /* Check Mode. */
        if (grub_strncmp(cur, normalModeString, mode_field_size) != 0)
        {
            cur = getNextString(&configFile);
            continue;
        }
        cur += mode_field_size;

        /* Get CBA core range indicated for this line (values are inclusive) */
        core_min = grub_strtoul(cur, &cur, 10);
        core_max = grub_strtoul(cur, &cur, 10);

        /* We've found core_max that we were looking for. */
        if ((core_min <= cbaCoresReq) && (core_max >= cbaCoresReq))
            break;

        cur = getNextString(&configFile);
    }

    /* if we did NOT find a corresponding entry for the given requested number of CBA cores
     * then there will be no CBA cores user, and core_max should be returned as 0.
     */
    configFileFree(&configFile);
    return core_max;
}


/**
 * Execute cpuid instruction. Input values are in eax and ecx registers. Output values are in eax,
 * ebx and ecx registers.
 */
static inline void
nativeCpuId
    (
    unsigned int *eax,
    unsigned int *ebx,
    unsigned int *ecx,
    unsigned int *edx
    )
{
    __asm__ volatile("cpuid"
        : "=a" (*eax), "=b" (*ebx), "=c" (*ecx), "=d" (*edx)
        : "0" (*eax), "2" (*ecx)
        : "memory");
}

/**
 * Execute cpuid instruction to get total number of logical cores.
 * NOTE - This function returns total number of logical cores even when hyperthreading is
 * disabled in the BIOS.
 */
static unsigned int
getNumLogicalCores
    (
    void
    )
{

    unsigned int eax = CPUID_LEAF;
    unsigned int ebx = 0;
    unsigned int ecx = 0;
    unsigned int edx = 0;
    unsigned int numCores = 0;

    /* Set input registers eax to 0xB and ecx to 0. Bits [7:0] of ecx contain level number.
     * CPUID instruction returns number of logical cores in bits [15:0] of ebx register at
     * the given level. Increment level number in ecx and loop until cpuid returns 0 in ebx.
     */
    while (1)
    {
        nativeCpuId(&eax, &ebx, &ecx, &edx);
        if (ebx == 0)
            break;
        numCores = ebx & NUM_LOGICAL_CORES_MASK;
        ecx++;
        ecx &= LEVEL_NUM_MASK;
        eax = CPUID_LEAF;
    }

    return numCores;
}

/**
 * The guts of the 'cpu_map' command.
 * Determine the platform and set isolcpus.
 */
static grub_err_t
grub_cmd_cpu_map
    (
    grub_command_t cmd __attribute__ ((unused)),
    int argc,
    char *argv[]
    )
{
    char *varName;
    char *cbaReqStr;
    unsigned int numCores;
    unsigned int cbaCores;
    char buf[50];

    if (argc == 0 || argc > 2)
    {
        return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("one or two arguments expected"));
    }

    /* If "DEBUG" env var exists, enable prints */
    enableDbgPrints = grub_env_get("DEBUG") ? 1 : 0;

    varName = argv[0];
    DBG_PRT("variable: %s\n", varName);

    if (argc == 2)
    {
        cbaReqStr = argv[1];
        DBG_PRT("CBA req: %s\n", cbaReqStr);
    }
    else
    {
        cbaReqStr = (char*) "0";
        DBG_PRT("CBA def: %s\n", cbaReqStr);
    }
    cbaCores = grub_strtoul(cbaReqStr, NULL, 10);

    numCores = getNumLogicalCores();
    if (cbaCores != 0)
    {
        cbaCores = getCpuConfiguration(cbaCores);
        DBG_PRT("Updated cbaCores to: %u\n", cbaCores);
    }

    /* Note: cpuid instruction will not return correct number of logical cores when
     * hyperthreading is disabled. This code will still isolate the correct list of cores.
     * For e.g. 4-core CPU with HT off: numCores = 8. isolcpus=1-3,5-7. As only cores 0-3 are
     * online, isolcpus will correctly isolate cores 1-3, and ignore 5-7.
     * This code assumes that all our processors support hyperthreading. It will need to be
     * changed if we come across a processor that doesn't support hyperthreading.
     */
    if (numCores < 2)
        grub_snprintf(buf, sizeof(buf), " ");
    else if (numCores == 2)
        grub_snprintf(buf, sizeof(buf), "isolcpus=1");
    else if (numCores == 4)
        grub_snprintf(buf, sizeof(buf), "isolcpus=1,3");
    else
        grub_snprintf(buf, sizeof(buf), "isolcpus=%u-%u,%u-%u",
            1, numCores / 2 - (1 + cbaCores), numCores / 2 + 1, numCores - (1 + cbaCores));

    DBG_PRT("Set %s=%s\n", varName, buf);
    grub_env_set(varName, buf);

    return GRUB_ERR_NONE;
}

/* Register/unregister our command with module init/exit */
static grub_command_t cmd;

GRUB_MOD_INIT(cpu_map)
{
    cmd = grub_register_command("cpu_map", grub_cmd_cpu_map, N_("ENV_VAR"),
              N_("Sets isolcpus kernel config parameter to isolate CPU cores from scheduler."));
}

GRUB_MOD_FINI(cpu_map)
{
    if (cmd)
    {
        grub_unregister_command(cmd);
        cmd = 0;
    }
}
