/*******************************************************************************

   read_env.c
   Copyright 2019 NetApp, Inc.

DESCRIPTION: Grub2 module to read in a file and set the environment variables
base on the content.   Each line will be ENVVAR=value format

NOTES:
Environment file is defined as follows:
  Lines beginning with '#' are comments.
  All other lines must contain 2 values, separated by "=".
    The first value is the environment variable name
    The second value is the environment variable value

LICENSE:
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/


/***  INCLUDES  ***/
#include <config.h>
#include <config-util.h>

#include <grub/dl.h>
#include <grub/env.h>
#include <grub/file.h>
#include <grub/misc.h>
#include <grub/command.h>

/***  CONSTANT DEFINITIONS  ***/
GRUB_MOD_LICENSE ("GPLv3+");
GRUB_MOD_DUAL_LICENSE ("All Rights Reserved");


/***  MACRO DEFINITIONS  ***/
#define ENV_FILE_NAME    "grub.env"
#define ENV_FILE         "(lvm/system-raid.active)/conf/" ENV_FILE_NAME

#define DBG_PRT(arg...)  do{ if(enableDbgPrints) grub_printf(arg); } while (0)

/***  TYPE DEFINITIONS  ***/
typedef struct {
    char*       start;
    char*       end;
    char*       cur;
    grub_off_t  len;
} EnvFile;

/***  LOCALS  ***/
static int enableDbgPrints=0;

/***  PROCEDURES  ***/

/**
 * Takes a pointer and returns a pointer to the next new line (or NULL if there
 * are no new lines remaining).
 */
static char*
getNextString(EnvFile* envFile)
{
    char* cur = envFile->cur + grub_strlen(envFile->cur);

    while (cur < envFile->end && *cur == '\0')
        cur++;

    envFile->cur = cur;

    if (cur < envFile->end)
        return cur;

    return 0;
}

/**
 * Free up memory allocated when file was open and read
 */
static void
envFileFree(EnvFile* envFile)
{
    grub_free(envFile->start);
    envFile->start = 0;
}


/**
 * Read in the entire environment file and store in structure.
 * Also convert all newlines to string terminators so that it is easier to parse.
 */
static int
readEnvFile(EnvFile* envFile)
{
    grub_file_t  file;
    char        *start;

    file = grub_file_open(ENV_FILE);
    if (!file)
    {
        /* Clear error so that is does not fail further processing */
        grub_errno = 0;

        DBG_PRT("Could not open env file: %s\n", ENV_FILE);
        return 0;
    }

    envFile->start = (char *)grub_zalloc(file->size);
    if (!envFile->start)
    {
        grub_printf("read_env: Could not allocate enough memory\n");
        grub_file_close(file);
        return 0;
    }

    envFile->len = grub_file_read(file, envFile->start, file->size);
    if (envFile->len != file->size)
    {
        grub_printf("read_env: Read data did not equal size of file\n");
        envFileFree(envFile);
        grub_file_close(file);
        return 0;
    }
    grub_file_close(file);

    envFile->cur = envFile->start;
    envFile->end = envFile->start + envFile->len;
    start = envFile->start;

    /* Scrub the file to get rid of newlines, replacing them with null terminator so that */
    /* parsing is easier */
    while (start < envFile->end)
    {
        if (*start == '\r' || *start == '\n')
            *start = '\0';
        start++;
    }

    return 1;
}


/**
 * Read in the environment file and post each entry.
 */
static void
postEnvVariables(void)
{
    EnvFile envFile;
    char* cur;
    char* allocPtr;
    char* envName = 0;

    /* If cannot read the environment file, then simply return.
     * Grub cfg file should have default values applied.
     */
    if (!readEnvFile(&envFile))
        return;

    cur = envFile.cur;

    while (cur != NULL && *cur != '\0')
    {
        /* First make a duplicate of the string since it will be modified */
        /* keeping allocated address for freeing later */
        allocPtr = grub_strdup(cur);
        cur = allocPtr;

        /* Advance to the next non-white space character */
        while((*cur != '\0') && grub_isspace(*cur))
            cur++;

        /* If comment line, then skip to next line in file */
        if ( *cur == '#')
        {
            cur = getNextString(&envFile);
            grub_free(allocPtr);
            continue;
        }

        /* 'cur' now points to the start of the environmental variable name */
        envName = cur;

        /* Search for the '=' delimiter */
        while ((*cur != '\0') && (*cur != '='))
            ++cur;

        /* If 'cur' points to '=' then we found an environmental variable assignment.
         * Split the string into 2 parts by setting '=' to '\0'.
         * Then take the second part as the assignment value and set the pair
         * as environment variables
         */
        if (*cur == '=')
        {
            *cur = '\0';
            ++cur;
            grub_env_set(envName, cur);
            DBG_PRT("Set %s=%s\n", envName, cur);
        }

        grub_free(allocPtr);
        cur = getNextString(&envFile);
    }

    envFileFree(&envFile);
}

/**
 * The guts of the 'read_env' command.
 */
static grub_err_t
grub_cmd_read_env (grub_command_t cmd __attribute__ ((unused)),
                   int argc __attribute__ ((unused)),
                   char *argv[] __attribute__ ((unused)))
{
    /* If "DEBUG" env var exists, enable prints */
    enableDbgPrints = grub_env_get("DEBUG") ? 1 : 0;

    /* Read in the environment variables and post them */
    postEnvVariables();

    DBG_PRT("read_env end\n");
    return GRUB_ERR_NONE;
}


/* Register/unregister our command with module init/exit */
static grub_command_t cmd;

GRUB_MOD_INIT(read_env)
{
    cmd = grub_register_command("read_env", grub_cmd_read_env, N_(""),
                                N_("Reads in environment variables from file and posts them for usage by grub"));
}

GRUB_MOD_FINI(read_env)
{
    if (cmd)
    {
        grub_unregister_command(cmd);
        cmd = 0;
    }
}

